unit Visitor;

interface

uses
  CustomVisitor, DabBoozePlant, DabDepartment, DabEmployee, Classes;

type
  // All visitor subclasses for the Dab classes have this form. However, there
  // is a bit of a problem. The Visit methods need to refer to the class types
  // of the elements being visited, and the element classes need to refer to
  // the base Visitor class in their Accept methods.
  // We could avoid this dilemma by using an Iterator, or by cheating a bit and
  // putting the base Visitor class in the unit where all the possible elements
  // are defined. This is a bit ugly though. So my solution is to use a
  // factory method in another class.
  // This class should not be instantiated.
  TCustomDabVisitor = class(TCustomVisitor)
  protected
    // The GoF suggest using names like:
    //    VisitElementA
    //    VisitElementB
    // for the visitor methods for each object structure element class
    //
    // In Delphi we can overload methods, which I think is neater, especially
    // in the traversal code in the object structure classes
    //
    // Visit methods are virtual so that visitor descendants can choose
    // which methods to implement - it may not be necessary to implement them
    // all for every operation
    procedure Visit(Element : TDabBoozePlants); overload; virtual;
    procedure Visit(Element : TDabBoozePlant); overload; virtual;
    procedure Visit(Element : TDabDepartments); overload; virtual;
    procedure Visit(Element : TDabDepartment); overload; virtual;
    procedure Visit(Element : TDabEmployees); overload; virtual;
    procedure Visit(Element : TDabEmployee); overload; virtual;
  public
    // This is the factory method
    procedure Visit(Element : TObject); overload; override;
  end;


  // A concrete visitor class to pretty print names
  TListEmployees = class(TCustomDabVisitor)
  private
    // Visitors can accumulate/track state
    FList : TStrings;

    procedure AddString(const AStr : string;Indent : Integer);
  public
    constructor Create(AList : TStrings);

    // Only need to do something for these methods, as the other collections
    // have no information we want to print
    procedure Visit(Element : TDabBoozePlant); override;
    procedure Visit(Element : TDabDepartment); override;
    procedure Visit(Element : TDabEmployee); override;
  end;


  // A concrete visitor class to list pay amounts
  TListPayAmounts = class(TCustomDabVisitor)
  private
    FList           : TStrings;
    FTotalPayAmount : Double;
  public
    constructor Create(AList : TStrings);

    procedure Visit(Element : TDabBoozePlants); override;
    procedure Visit(Element : TDabEmployee); override;

    property TotalPayAmount : Double read FTotalPayAmount;
  end;


implementation

uses
  SysUtils;
  
{ TCustomDabVisitor }

procedure TCustomDabVisitor.Visit(Element : TDabBoozePlants);
begin
end;


procedure TCustomDabVisitor.Visit(Element : TDabBoozePlant);
begin
end;


procedure TCustomDabVisitor.Visit(Element : TDabDepartments);
begin
end;


procedure TCustomDabVisitor.Visit(Element : TDabDepartment);
begin
end;


procedure TCustomDabVisitor.Visit(Element : TDabEmployees);
begin
end;


procedure TCustomDabVisitor.Visit(Element : TDabEmployee);
begin
end;


procedure TCustomDabVisitor.Visit(Element : TObject);
begin
  if Element is TDabBoozePlants then begin
    Visit(TDabBoozePlants(Element));
  end else if Element is TDabBoozePlant then begin
    Visit(TDabBoozePlant(Element));
  end else if Element is TDabDepartments then begin
    Visit(TDabDepartments(Element));
  end else if Element is TDabDepartment then begin
    Visit(TDabDepartment(Element));
  end else if Element is TDabEmployees then begin
    Visit(TDabEmployees(Element));
  end else if Element is TDabEmployee then begin
    Visit(TDabEmployee(Element));
  end;
end;


{ TListEmployees }

const
  IndentAmount = 6;

constructor TListEmployees.Create(AList : TStrings);
begin
  inherited Create;
  FList := AList;
end;


procedure TListEmployees.AddString(const AStr : string;Indent : Integer);
begin
  FList.Add(StringOfChar(' ',Indent) + AStr);
end;


procedure TListEmployees.Visit(Element : TDabBoozePlant);
begin
  AddString(Element.PlantName + ',  ' + Element.Location,0);
end;


procedure TListEmployees.Visit(Element : TDabDepartment);
begin
  AddString(Element.DepartmentName,IndentAmount);
end;


procedure TListEmployees.Visit(Element : TDabEmployee);
  var
    TempIndent : Integer;
begin
  if Element.Collection.Owner is TDabBoozePlant then begin
    TempIndent := IndentAmount;
  end else begin
    TempIndent := IndentAmount * 2;
  end;

  AddString(Element.FullName,TempIndent);
end;


{ TListPayAmounts }

constructor TListPayAmounts.Create(AList : TStrings);
begin
  inherited Create;
  FList := AList;
end;


procedure TListPayAmounts.Visit(Element: TDabBoozePlants);
begin
  FTotalPayAmount := 0;
end;


procedure TListPayAmounts.Visit(Element : TDabEmployee);
begin
  FTotalPayAmount := FTotalPayAmount + Element.PayAmount;
  FList.Add(Format('%s   %f',[Element.FullName,Element.PayAmount]));
end;


end.
